/*************************************************************************
Title:    Temperaturlogger fr acht anloge Sensoren (KTY 81-110)
		  (Ausgabe ber LCD und USART in unterschiedlichen Zeitintervallen 
		  incl. kleinem UART-Menu);
		  Speicherung de Daten auf einem externen EEPROM (24C512);
		  Zeitraum der Speicherintervalle variabel einstellbar;
		  Speicherung und Ausgabe der Maximaltemperaturwerte;
		  Tastendruck lscht Max.temp.werte;
		  Umschaltung der LCD-Anzeige per software-entprellter Taste zur Darstellung der max. Temperatur
		  UART-Sendefortschritt der gespeicherten Daten auf dem LCD ablesbar;
		  Funktionsfhig!! 06.3.10 
**************************************************************************/
#define F_CPU 3276800L
#include <stdlib.h>
#include <avr/io.h>
#include <inttypes.h>
#include <string.h>
#include <avr/pgmspace.h> 
#include <util/delay.h>
#include "lcd.h"
#include <avr/interrupt.h>
#include <avr/iom32.h>
#include "24c64.h"

#ifndef F_CPU

#warning "F_CPU war noch nicht definiert, wird nun nachgeholt mit 4000000"
#define F_CPU 4000000UL  // Systemtakt in Hz - Definition als unsigned long beachten 
                         // Ohne ergeben sich unten Fehler in der Berechnung
#endif
 
#define BAUD 4800UL      // Baudrate 
					//Gibt Fehlermeldung aus, falls Fehlerrate bei dieser Baudrate zu gro ist
 
// Berechnungen
#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)   // clever runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))     // Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD) // Fehler in Promille, 1000 = kein Fehler.
 
#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #error Systematischer Fehler der Baudrate grsser 1% und damit zu hoch! 
#endif


/* Programm-Parameter: */
#define scroll_time   	2		//1 Einheit ^= Zeilensprung im LCD jede sec
#define update_time	 	4		//1 Einheit ^= Aktualisierung der Temp.messwerte jede sec
#define write_eep_time	120		//120 = alle 2 Minuten (1 Einheit ^= Abspeichern der aktuellen Temp.Messwerte jede sec
#define eeprom_cap	 	65536	//Speichergre EEPROM in Byte !
#define lcd_light_time	60		//automatisches Ausschalten des LCD-Backlights (1 Einheit = 1 sec Leuchtdauer)
 
#define KEY_DDR         DDRB
#define KEY_PORT        PORTB
#define KEY_PIN         PINB
#define KEY0            0
#define KEY1            1
#define KEY2            2
#define ALL_KEYS        (1<<KEY0 | 1<<KEY1 | 1<<KEY2)
 
#define REPEAT_MASK     (1<<KEY1 | 1<<KEY2)       // repeat: key1, key2
#define REPEAT_START    50                        // after 500ms
#define REPEAT_NEXT     20                        // every 200ms
 
#define LED_DDR         DDRC
#define LED_PORT        PORTC
#define LED0            3
#define LED1            2
#define LED2            4

const char Sensor1[] PROGMEM = "Solar VL:     ";
const char Sensor2[] PROGMEM = "Solar RL:     ";
const char Sensor3[] PROGMEM = "WWTem. oben:  ";
const char Sensor4[] PROGMEM = "WWTem. Mitte: ";
const char Sensor5[] PROGMEM = "WWTem. unten: ";
const char Sensor6[] PROGMEM = "WWPumpe VL:   ";
const char Sensor7[] PROGMEM = "WWPumpe RL:   ";
const char Sensor8[] PROGMEM = "Heizung VL:   "; 

const char* Bezeichnung[] PROGMEM = { Sensor1, Sensor2, Sensor3, Sensor4,
									  Sensor5, Sensor6, Sensor7, Sensor8 };

volatile uint8_t line, buffer, flag_uart, flag_temp,
				 counter_update_temp, counter_scroll_lcd, 
				 temp[8], temphigh[8],counter_lcd_light, sekunde, minute;
volatile uint16_t adcbit, adctemp;
volatile uint32_t  counter_write_eeprom, byte_counter;

volatile uint8_t key_state;                                // debounced and inverted key state:
                                                  		   // bit = 1: key pressed
volatile uint8_t key_press;                                // key press detect
 
volatile uint8_t key_rpt;                                  // key long press and repeat

/*
** function prototypes
*/ 


void init_timer1_ctc()
{
    //Init des Timer1 CTC-Modus
	TCCR1A |= (0<<COM1A1)|(0<<COM1A0)|(0<<COM1B1)| (0<<COM1B0);//no hardware-triggering
	TCCR1A |= (0<<WGM11) |(0<<WGM10);				//set Timer1 to CTC-Mode
	TCCR1B |= (1<<WGM12) |(1<<CS12)  |(0<<CS11)  |(1<<CS10);	//set prescaler to /1024
	TIMSK  |= (1<<OCIE1A);				//set out compare match a interrupt enable
				//-> bei 3,2768 MHz jede Sek. auslsen der ISR
	OCR1A=3200-1; //erreicht Timer in diesen Wert kommt es zum Auslsen der ISR-Funktion ISR (TIMER1_COMPA_vect)
	


}

void init_timer0_ovf()
{ 
  TCCR0 |= (1<<CS02)|(1<<CS00);			// divide by 1024
  TIMSK |= 1<<TOIE0;				// enable timer interrupt
}



// USART initialisieren
void uart_init(void){
	// Baudrate einstellen (Normaler Modus)
	UBRRH = UBRR_VAL >> 8;
    UBRRL = UBRR_VAL & 0xFF;

	// Aktivieren von receiver, transmitter und interrupt (UART)
	UCSRB = (1<<TXEN) | (1<<RXEN) | (1<<RXCIE);

	// Einstellen des Datenformats: 8 Datenbits, 1 Stoppbit

	UCSRC |= (1<<URSEL)|(3<<UCSZ0);
}

uint16_t ReadChannel(uint8_t mux)
{
  uint8_t i;
  uint16_t result;
  cli();
  ADMUX = mux;                      // Kanal waehlen
  ADMUX |= (1<<REFS0);     			//externe Referenzspannung nutzen
  //ADMUX |= (1<<REFS1) | (1<<REFS0); // interne Referenzspannung nutzen
 
  ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS0);    // Frequenzvorteiler 
                               // setzen auf 8 (1) und ADC aktivieren (1)
 
  /* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
     also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
  ADCSRA |= (1<<ADSC);              // eine ADC-Wandlung 
  while ( ADCSRA & (1<<ADSC) ) {
     ;     // auf Abschluss der Konvertierung warten 
  }
  result = ADCW;  // ADCW muss einmal gelesen werden,
                  // sonst wird Ergebnis der nchsten Wandlung
                  // nicht bernommen.
 
  /* Eigentliche Messung - Mittelwert aus 4 aufeinanderfolgenden Wandlungen */
  result = 0; 
  for( i=0; i<4; i++ )
  {
    ADCSRA |= (1<<ADSC);            // eine Wandlung "single conversion"
    while ( ADCSRA & (1<<ADSC) ) {
      ;   // auf Abschluss der Konvertierung warten
    }
    result += ADCW;		    // Wandlungsergebnisse aufaddieren
  }
  ADCSRA &= ~(1<<ADEN);             // ADC deaktivieren (2)
 
  result /= 4;                     // Summe durch vier teilen = arithm. Mittelwert

  return result;
}

uint16_t umrechnung (uint16_t adc)
{
uint32_t y=0;
uint16_t z;
	
	for (z=1;z<=633;z++) y = (adcbit+y); //Multiplikation mit 633!! (ber entsprechende Addition)
	y-=150000; //mglicher Offset!!!
	y/=1000;
	return (y);
}

void uart_putc(unsigned char c)
{
    while (!(UCSRA & (1<<UDRE)))  /* warten bis Senden moeglich */
    {
    }                             
 
    UDR = c;                      /* sende Zeichen */
    //return 0;
}
 
 
/* puts ist unabhaengig vom Controllertyp */
void uart_puts (char *s)
{
    while (*s)
    {   /* so lange *s != '\0' also ungleich dem "String-Endezeichen" */
        uart_putc(*s);
        s++;
    }
}

void update_temp(){
uint8_t i;
i=0;

	for (i=0; i<8;i++){
		adcbit=ReadChannel(i);
		adctemp=umrechnung(adcbit);
		temp[i]=adctemp;
		if (temp[i]>temphigh[i]) temphigh[i]=temp[i]; //Hchsttemp.speicherung
													  
	}
}

void write_eeprom_bl()
{
uint8_t i;
	if(byte_counter<=(eeprom_cap-1))   //Angabe des Speicherplatzes (Bytes), danach wird
	{					  			// weiteres Schreiben in den EEPROM unterbunden.
		PORTC |= (1<<PC5);
		for (i=0;i<8;i++)
			{
			EEWriteByte(byte_counter, temp[i]);
			byte_counter++;
			}

		PORTC &= ~(1<<PC5);
	}
		if (byte_counter>=(eeprom_cap-1)) PORTC |= (1<<PC3);
}

void lcd_ausgabe_temp(void){

char Ergebnis[20];
char  buf[20];

	if(line<=3){
		lcd_gotoxy(0,line);
		lcd_puts("                    ");
		lcd_gotoxy(0,line);
		itoa(temp[line], Ergebnis,10);
		strcpy_P( buf, (const char*)( pgm_read_word( &(Bezeichnung[line]) ) ) );
		lcd_puts(buf);
		lcd_puts(Ergebnis);
		lcd_putc(0b10000000);
		lcd_puts("C");
	}

	else{
		lcd_gotoxy(0,line-4);
		lcd_puts("                    ");
		lcd_gotoxy(0,line-4);
		itoa(temp[line], Ergebnis,10);
		strcpy_P( buf, (const char*)( pgm_read_word( &(Bezeichnung[line]) ) ) );
		lcd_puts(buf);
		lcd_puts(Ergebnis);
		lcd_putc(0b10000000);
		lcd_puts("C");
	}

}

void lcd_ausgabe_temphigh(void){

char Ergebnis[20];
char  buf[20];
	lcd_gotoxy(0,0);
	lcd_puts("Max. Temperaturwerte");
	if(line<=2){
		lcd_gotoxy(0,line+1);
		lcd_puts("                    ");
		lcd_gotoxy(0,line+1);
		itoa(temphigh[line], Ergebnis,10);
		strcpy_P( buf, (const char*)( pgm_read_word( &(Bezeichnung[line]) ) ) );
		lcd_puts(buf);
		lcd_puts(Ergebnis);
		lcd_putc(0b10000000);
		lcd_puts("C");
	}

	else if(line>2 && line<=5) {
		lcd_gotoxy(0,line-2);
		lcd_puts("                    ");
		lcd_gotoxy(0,line-2);
		itoa(temphigh[line], Ergebnis,10);
		strcpy_P( buf, (const char*)( pgm_read_word( &(Bezeichnung[line]) ) ) );
		lcd_puts(buf);
		lcd_puts(Ergebnis);
		lcd_putc(0b10000000);
		lcd_puts("C");
	}
	else {
		lcd_gotoxy(0,line-5);
		lcd_puts("                    ");
		lcd_gotoxy(0,line-5);
		itoa(temphigh[line], Ergebnis,10);
		strcpy_P( buf, (const char*)( pgm_read_word( &(Bezeichnung[line]) ) ) );
		lcd_puts(buf);
		lcd_puts(Ergebnis);
		lcd_putc(0b10000000);
		lcd_puts("C");
	}

}

void erase_temphigh()
{
  uint8_t z;

  for (z=0;z<=7;z++)
  {	
  	temphigh[z]=1;
  }
}


uint8_t get_key_press( uint8_t key_mask )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}

uint8_t get_key_rpt( uint8_t key_mask )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_rpt;                            // read key(s)
  key_rpt ^= key_mask;                            // clear key(s)
  sei();
  return key_mask;
}

uint8_t get_key_short( uint8_t key_mask )
{
  cli();                                          // read key state and key press atomic !
  return get_key_press( ~key_state & key_mask );
}

uint8_t get_key_long( uint8_t key_mask )
{
  return get_key_press( get_key_rpt( key_mask ));
}
 
void check_uart(){
  //PORTC|=(1<<PC4);
  while ( !(UCSRA & (1<<RXC)) ); //warten bis Daten empfangen wurden
  
  buffer=UDR; 					 // Empfangsregister auslesen

}

void uart_print()
{
char buf[11],j;
uint32_t i;
uint8_t tempbuf;
j=1;
  PORTC |= (1<<PC5);
  for (i=0; i<=(byte_counter-1); i++)
  {
  	lcd_gotoxy(0,1);
  	lcd_puts("Status:\n");			//Ausgabe des Sendefortschritts auf dem LCD
  	ultoa(i,buf,10);
  	lcd_puts(buf);
  	lcd_puts(" von ");
  	ultoa(byte_counter,buf,10);
  	lcd_puts(buf);
  	lcd_puts("\nDaten gesendet");
 
  	if(j<=7)
	{
		tempbuf = EEReadByte(i); 
								
		if (tempbuf<100)	//berprfung, ob Temp.wert 2- oder 3-stellig ist.
							//Wichtig, um die Daten direkt in Excel einlesen zu knnen!
		{
			uart_puts(" S");
  			itoa(j,buf,10);
  			strcat(buf, " ");
  			uart_puts(buf);
			uart_puts(" ");
			itoa(tempbuf, buf,10);
			uart_puts(buf);
		}
		else 
		{
			uart_puts(" S");
  			itoa(j,buf,10);
  			strcat(buf, " ");
  			uart_puts(buf);
			itoa(tempbuf, buf,10);
			uart_puts(buf);
		}
			j++;
	}
	else 				//nach der Ausgabe des 8. Temp.wertes soll ein Zeilenumbruch
						//bei der Ausgabe ber den UART ausgegeben werden.
	{
		tempbuf=EEReadByte(i);

		if (tempbuf<100){
			uart_puts(" S");
  			itoa(j,buf,10);
  			strcat(buf, " ");
  			uart_puts(buf);
			uart_puts(" ");
			itoa(tempbuf, buf,10);
			uart_puts(buf);
			uart_puts("\n");
			j=1;
		}
		else {
			uart_puts(" S");
  			itoa(j,buf,10);
  			strcat(buf, " ");
  			uart_puts(buf);
			itoa(tempbuf, buf,10);
			uart_puts(buf);
			uart_puts("\n");
			j=1;
		}
	}
  }
  PORTC &=~ (1<<PC5);
  lcd_clrscr();
}	

// Timer1 CTC-Modus 
ISR (TIMER1_COMPA_vect){


	counter_write_eeprom++;
	counter_update_temp++;
	counter_scroll_lcd++;
	counter_lcd_light++;
	sekunde++;
		
}

// Timer/Counter0 Overflow 
ISR( TIMER0_OVF_vect )                            // every 10ms
{
  static uint8_t ct0, ct1, rpt;
  uint8_t i;
 
  TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms
 
  i = key_state ^ ~KEY_PIN;                       // key changed ?
  ct0 = ~( ct0 & i );                             // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                          // reset or count ct1
  i &= ct0 & ct1;                                 // count until roll over ?
  key_state ^= i;                                 // then toggle debounced state
  key_press |= key_state & i;                     // 0->1: key press detect
 
  if( (key_state & REPEAT_MASK) == 0 )            // check repeat function
     rpt = REPEAT_START;                          // start delay
  if( --rpt == 0 ){
    rpt = REPEAT_NEXT;                            // repeat delay
    key_rpt |= key_state & REPEAT_MASK;
  }
}
 

ISR(USART_RXC_vect) //UART-Men wird aktiviert, sobald ein Zeichen vom PC aus gesendet wurde.
{
    
    flag_uart=1;

}


int main(void)
{

	KEY_DDR &= ~ALL_KEYS;                // konfigure key port for input
  	KEY_PORT |= ALL_KEYS;                // and turn on pull up resistors
 	
	LED_DDR |= (1 << DDC2) | (1<<DDC3) | (1<<DDC4) | (1<<DDC5) | (1<<DDC6);       /* Pin PB1 output              */
    PORTC |= (1<<PC6); //Hintergrundbeleuchtung anschalten
	
	// initialize display, cursor off 
   	lcd_init(LCD_DISP_ON);
	lcd_clrscr();
	lcd_gotoxy(0,0);
	lcd_puts("Initialisierung...");

	PORTC |= (1<<PC2); 
	_delay_ms(500);
	PORTC &= ~(1<<PC2); 
	_delay_ms(500);
	PORTC |= (1<<PC4); 
	_delay_ms(500);
	PORTC &=~ (1<<PC4); 
	_delay_ms(500);
	PORTC |= (1<<PC5); 
	_delay_ms(500);
	PORTC &= ~(1<<PC5); 
	_delay_ms(500);
	PORTC |= (1<<PC3); 
	_delay_ms(500);
	PORTC &= ~(1<<PC3); 

	PORTC |= (1<<PC2);

    // initialize display, cursor off 
   	lcd_init(LCD_DISP_ON);
	lcd_clrscr();
	lcd_gotoxy(0,0);

	// EEPROM initialisieren
	EEOpen();

	//USART initialisieren
	uart_init();

	init_timer1_ctc();
  	init_timer0_ovf();

	line=0;
	byte_counter=0;
	flag_temp=1;
	sekunde=0;
	minute=0;
	counter_lcd_light=0;
	

	update_temp();
	write_eeprom_bl();
	sei();

	while (1) 
	{
	  /*Uhrfunktion
	  lcd_gotoxy(0,0);
	  itoa(minute,buf,10);
	  lcd_puts(buf);
	  lcd_puts(":");
	  itoa(sekunde,buf,10);
	  lcd_puts(buf);
	  if(sekunde==60) lcd_clrscr();
	  */	
	 	if(counter_update_temp>=update_time)	//Aktualisierung der Temp.werte
		{
			cli();	
			PORTC |= (1<<PC4);
			update_temp();
			counter_update_temp=0;
			PORTC &= ~(1<<PC4);
			sei();
		}

		if(counter_scroll_lcd>=scroll_time)		//Zeilensprnge auf dem LCD
		{
			cli();
			if(flag_temp!=1) lcd_ausgabe_temphigh();	//Zwischen Max. und aktuellen Temp.wert
														// whlen (ber Taster)
			else lcd_ausgabe_temp();
			line++;
			counter_scroll_lcd=0;
			if(line==8){
					line=0;
				}
			sei();
		}

		if (counter_lcd_light>=lcd_light_time)		//Ausschalten der Hintergrundbeleuchtung
		{
		  	PORTC &= ~(1<<PC6);
		}

		if (counter_write_eeprom>=write_eep_time){	//EEPROM-Schreibaufruf
			cli();
			write_eeprom_bl();
			counter_write_eeprom=0;
			sei();
		}

		if( get_key_short( 1<<KEY2 ))		//Taster kurz gedrckt
		{	
			cli();
			lcd_clrscr();
			line=0;
			flag_temp ^= 1;
			counter_lcd_light=0;
			PORTC |= (1<<PC6);				//Hintergrundbeleuchtung an
			sei();
		}

		if( get_key_long( 1<<KEY2 ))	//Taster lang gedrckt
		{
			lcd_clrscr();
			counter_lcd_light=0;
			PORTC |= (1<<PC6);				//Hintergrundbeleuchtung an
			lcd_puts("Max. Tempwerte \nzurueckgesetzt!"); 
			erase_temphigh();
		}
		 
		/* Uhrfuntion
		if (sekunde>=60)
		{
			minute++;
			sekunde=0;
		}
		*/

		if(flag_uart==1){ 		//UART-Men
			cli();
				
				lcd_clrscr();
				while (flag_uart==1)
				{
					PORTC |= (1<<PC6); //Hintergrundbeleuchtung anschalten
					lcd_gotoxy(0,0);
					lcd_puts("Daten (S)enden,\n(L)oeschen oder\nUART (v)erlassen");
					check_uart();
					if(buffer=='S'|| buffer=='s')
						{
							lcd_clrscr();
							lcd_puts("Senden.Bitte warten!\n");
							uart_print();
							buffer='0';
																			
						}
						if(buffer=='L'||buffer=='l')
						{
							lcd_clrscr();
							lcd_puts("Daten werden\n geloescht!\n Bitte warten!!");
							uart_puts("Daten werden geloescht! Bitte warten!!");
							//erase_eeprom(); Ein Erase wrde zu lange dauern, daher einfach
							// byte_counter null setzen. Damit werden die Alten einfach berschrieben.
							// Dies beeinflut die Ausgabe ber UART nicht.
							erase_temphigh();
							byte_counter=0;
							lcd_clrscr();
							lcd_puts("Daten geloescht!");
							uart_puts("\nDaten geloescht!");
							write_eeprom_bl();
							PORTC &=~(1<<PC3);
							lcd_clrscr();
							buffer='0';
							
							
						}
						if(buffer=='B'||buffer=='b'||buffer=='v'||buffer=='V')
						{	
							buffer='0';
							flag_uart=0;
							lcd_clrscr();
							
						}
				}
			
			sei();
		}	
			
	}
	

}



